/*

    File: advectFragment.shader

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

varying vec2 tc;
varying vec4 color;

#ifdef VERTEX_SHADER
void main() {
	color = gl_Color;
	tc = gl_MultiTexCoord0.xy;
	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
}
#endif

#ifdef FRAGMENT_SHADER

uniform sampler2D colorMap;
uniform sampler2D heightMap;

uniform float gradientStep;
uniform float advectStep;

//dirty hack for flipping the y coord for heightmap lookup and heightmap gradients.
//its cuz gl uses messed up flipped tex's and what not ...
//my only other fix would be to flip the y when loading the texture ...
uniform float flipHeightMap;

//here's the basis along which we advect our junk
uniform mat2 advectMatrix;

void main() {
//discard;
	//heres how it worked for dr zhang:
	//find the gradient of the height
	//rotate 90 degrees: x,y -> -y,x
	//advect along that vector
	//and thats your new texture coorindate

	vec4 cxp = texture2D(heightMap, vec2(tc.x + gradientStep, flipHeightMap*tc.y));
	vec4 cxn = texture2D(heightMap, vec2(tc.x - gradientStep, flipHeightMap*tc.y));
	vec4 cyp = texture2D(heightMap, vec2(tc.x, flipHeightMap*(tc.y + gradientStep)));
	vec4 cyn = texture2D(heightMap, vec2(tc.x, flipHeightMap*(tc.y - gradientStep)));
	
	//get our height at deltas
	vec3 grey = vec3(.3, .59, .11);
	float axp = dot(grey, cxp.xyz);
	float axn = dot(grey, cxn.xyz);
	float ayp = dot(grey, cyp.xyz);
	float ayn = dot(grey, cyn.xyz);
	//calc grad
	vec2 grad = vec2(axp - axn, ayp - ayn);

	//and trasnform tc accordingly
	vec2 newtc = tc + advectStep * normalize(advectMatrix * grad);
	gl_FragColor.rgb = texture2D(colorMap, newtc).rgb * color.rgb;
	gl_FragColor.a = color.a;
}
#endif
