--[[

    File: watercolor.lua

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

--]]

require 'common.glutil'
require 'common.shader'
require 'common.tex'

for k,v in pairs(gl) do _G[k] = v end

fboSize = 1024		-- width & height -- lets make it square for simplicity's sake

--[[ controls
pass the global names in here
i'm gonna to global references at first
but later we can just as well make them strings to eval()
lua is nice like that
so in a sense they will be getter/setter functions
when i do that i'll have to make a separate title parameter
i guess i could do default values too if i wanted, but those might as well be in a separate config area
--]]
glush.gui.slider('blendCoeff',			0, 1-7/100, 1)
glush.gui.checkbox('showBackground',	true)
glush.gui.checkbox('showCow',			true)
glush.gui.checkbox('advectByBackground', true)
glush.gui.checkbox('advectByAdvection',	false)
glush.gui.checkbox('flipAdvection',		true)
glush.gui.checkbox('twoStroke',			true)		-- do we do a two-cycle or a one-cycle (copy the dest back into the src) ?
glush.gui.slider('gradientStep',		-1, .0015, 1)
glush.gui.slider('objScale',	.1,	1,	10)
glush.gui.slider('advectStep',			-.1, .0015, .1)	-- exponential slider?
glush.gui.basis('advectMatrix',			{0, -1, 1, 0})

-- i need to combine these two
-- i need to create image objects ...
backgroundTex = Tex2D.load('watercolor/dandelions.jpg')
--backgroundTex = Tex2D.load('/Users/christophermoore/Desktop/n19716634_31813090_9346.jpg')
--backgroundTex = Tex2D.load('/Users/christophermoore/Pictures/Best Of/Dining Room Lamp.jpg')
--backgroundTex = Tex2D.load('/Users/christophermoore/Pictures/Best Of/Ladybugs of Wonder.jpg')
--backgroundTex = Tex2D.load('ruby.jpg')
--backgroundTex = Tex2D.load('autumn.jpg')
--backgroundTex = Tex2D.load('/Users/christophermoore/Pictures/2007-Kawasaki-Ninja250R-EX250d.jpg')

--glush.gui.tex2D('backgroundTex')

noiseTex = glush.createTex3D('watercolor/noise3d.064.tex')

-- i set up enable/disable to accept as many params as you can stuff in it
-- i know, i'm stretching my jurisdiction.
glEnable(GL_NORMALIZE, GL_CULL_FACE)

-- make your fbo
fboID = glGenFramebuffer()
glBindFramebuffer(GL_FRAMEBUFFER, fboID)

-- [[
-- make a depth buffer render target
-- do you need a depth target to do depth testing when rendering to a fbo?
fboDepthID = glGenRenderbuffer()
glBindRenderbuffer(GL_RENDERBUFFER, fboDepthID)
glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, fboSize, fboSize)
glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, fboDepthID)
glBindFramebuffer(GL_FRAMEBUFFER, 0)
--]]

-- make a tex 2D color buffer render target
-- make two of them.  two-state system.
fboTexCur = glGenTexture()		-- two for advecting, cant rw to the same place 
fboTexNext = glGenTexture()		-- so we need two
fboTexBase = glGenTexture()		-- a third for holding the static scene pre-processing

for i,v in ipairs({fboTexCur, fboTexNext, fboTexBase}) do
	glBindTexture(GL_TEXTURE_2D, v)
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, fboSize, fboSize, 0, GL_RGBA, GL_UNSIGNED_BYTE)
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
	-- autogen mipmapping isnt working for me up front
	--	glGenerateMipmap(GL_TEXTURE_2D)
	-- this is tho ... not really
	-- glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR)
	-- glTexParameter(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, true)
	-- glHint(GL_GENERATE_MIPMAP_HINT, GL_NICEST)
end

print('compiling cow.shader')
cowShader = ShaderProgram.new{
	vertexCode='#define VERTEX_SHADER\n'..io.readfile('watercolor/cow.shader');
	fragmentCode='#define FRAGMENT_SHADER\n'..io.readfile('watercolor/cow.shader');
	uniforms = {
		noiseTex = 0;
	};
}

print('compiling advectFragment.shader')
advectFragmentShader = ShaderProgram.new{
	vertexCode = '#define VERTEX_SHADER\n'..io.readfile('watercolor/advectFragment.shader');
	fragmentCode = '#define FRAGMENT_SHADER\n'..io.readfile('watercolor/advectFragment.shader');
	uniforms = {
		colorMap = 0;
		heightMap = 1;
		gradientStep = gradientStep;
		advectStep = advectStep;
		flipHeightMap = 1;
		advectMatrix = advectMatrix;
	}
}
	
-- turn off clear on update ...
glush.clearFlags(GL_DEPTH_BUFFER_BIT)

function drawScreenQuad(texYDir)
	glBegin(GL_QUADS)
	glTexCoord(0,0)			glVertex(0,0)
	glTexCoord(1,0)			glVertex(1,0)
	glTexCoord(1,texYDir)	glVertex(1,1)
	glTexCoord(0,texYDir)	glVertex(0,1)
	glEnd()
end

function drawScreenTex(tex, texYDir)
	glEnable(GL_TEXTURE_2D)
	glBindTexture(GL_TEXTURE_2D, tex)
	drawScreenQuad(texYDir)
	glDisable(GL_TEXTURE_2D)
end

-- so this is fragment-based advection
-- i'm gonna swap it out with vertex-based advection soon
-- and it will be cool
function advection()
	
	glEnable(GL_BLEND)
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
	glColor(1,1,1, blendCoeff)
	
	-- bind our program
	advectFragmentShader:use()
	-- update the uniforms
	glUniformf('gradientStep', gradientStep)
	glUniformf('advectStep', advectStep)
	glUniformMatrix('advectMatrix', advectMatrix)

	glActiveTexture(GL_TEXTURE0)	--texture0 is the colormap
	glEnable(GL_TEXTURE_2D)
	glBindTexture(GL_TEXTURE_2D, fboTexCur)

	glActiveTexture(GL_TEXTURE1)	--texture1 is the edgemap src
	glEnable(GL_TEXTURE_2D)
	-- OH NO
	-- we are advecting along the dandelion edges!
	-- not the scene edges at all
	-- oh my... 
	-- its not view-independent at all
	-- but its based on the background image
	local flipHeightMap = 1
	if advectByBackground then
		backgroundTex:bind()
		flipHeightMap = -flipHeightMap		-- because the backgroundTex is loaded upside down
	elseif advectByAdvection then
	-- or we can advect by the current advected texture
		glBindTexture(GL_TEXTURE_2D, fboTexCur)
	else
	-- to fix that we do this:
		glBindTexture(GL_TEXTURE_2D, fboTexBase)
	-- but it looks nowhere as good
	end
	-- user-controlled to simulate the original advection bug, which looked so cool
	if flipAdvection then flipHeightMap = -flipHeightMap end

	glUniformf('flipHeightMap', flipHeightMap)
	
	drawScreenQuad(1)
	
	glActiveTexture(GL_TEXTURE1)
	glDisable(GL_TEXTURE_2D)
	glActiveTexture(GL_TEXTURE0)
	glDisable(GL_TEXTURE_2D)
	
	glUseProgramObject(0)
	glDisable(GL_BLEND)
	
	glColor(1,1,1,1)
end

function drawScene()
	--clear the buffer
	glClear(GL_COLOR_BUFFER_BIT)
	
	if showBackground then
		-- draw our background
		drawOverlay(function()
			drawScreenTex(backgroundTex.id, -1) 
		end)
	end
	
	-- and draw the cow over it
	if showCow then
		-- clear the depth buffer
		glClear(GL_DEPTH_BUFFER_BIT)
		
		glEnable(GL_DEPTH_TEST)
		cowShader:use()
		glScale(objScale,objScale,objScale)
		glush.drawOBJ('cow.obj')
		glUseProgramObject(0)
		glDisable(GL_DEPTH_TEST)
	end
end

function drawToFboTex2D(fbo, tex, callback)
	glBindFramebuffer(GL_FRAMEBUFFER, fbo)
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, tex, 0)
	local fboStatus = glCheckFramebufferStatus(GL_FRAMEBUFFER)
	if fboStatus ~= GL_FRAMEBUFFER_COMPLETE then
		print('got a bad fboStatus: '..fboStatus)
		return
	end
	
	callback()
	
	glBindFramebuffer(GL_FRAMEBUFFER, 0)
end

-- callback from the main cheese
-- here's a thought
-- how about a seperate callback for whether you want to
-- display inside or outside the glClear / matrix transform routines?
function glush.display()
--[[
the watercolor gameplan
1. draw the cow
2. save it to a fbo
3. render the fbo to the whole screen
BUT WITH A SHADER!!

so how do we do accumulation?
--]]

	-- update the fbo	
	do
		glPushAttrib(GL_VIEWPORT_BIT)
		glViewport(0,0,fboSize,fboSize)
		--you dont need to reset the projection do you?
		
		-- first render the ordinary scene to the third fbo tex
		-- hey do you need more than one fbo, or can you just keep rebinding it?
		-- so long as all the texs have the same size?
		drawToFboTex2D(fboID, fboTexBase, function()
			-- render the static scene to the third fbo tex
			drawScene()
		end)
		
		drawToFboTex2D(fboID, fboTexNext, function()
			-- draw scene here
			drawOverlay(function()
				glClear(GL_COLOR_BUFFER_BIT)
				
				-- start with the base scene
				drawScreenTex(fboTexBase,1)

				-- run the advection
				advection()
			end)
		end)

		
		--swap current and next
		--if i was cool i would copy from next back to current
		if twoStroke then
			local tmp
			tmp = fboTexCur
			fboTexCur = fboTexNext
			fboTexNext = tmp
		else
			-- copy the next tex into the current tex
			drawToFboTex2D(fboID, fboTexCur, function()
				drawOverlay(function()
					glClearColor(GL_COLOR_BUFFER_BIT)
					drawScreenTex(fboTexNext,1)
				end)
			end)
		end

		-- restore viewport
		glPopAttrib()
	end
	
	if dumpTheScreen then
		writeTex2D(fboTexCur)
		dumpTheScreen = nil
	else
		drawOverlay(function()
			drawScreenTex(fboTexCur,1) 
		end)
	end
	
	local err = glGetError()
	if err ~= 0 then print('glGetError: ' ..err) end
end
