--[[

    File: watercolor.v1.lua

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

--]]

-- width & height -- lets make it square for simplicity's sake
fboSize = 1024
blendCoeff = (1 - 3/100)

math.fract = function(x) return x - math.floor(x) end

function defines(...)
	local lines = {...}
	if #lines == 1 and type(lines[1]) == 'table' then
		print('defines switching to table mode')
		lines = lines[1] 
	end
	local str = ''
	for i,line in pairs(lines) do
		str = str .. '#define ' ..line .. '\n'
	end
	return str
end

backgroundTex = createTex2D('dandelions.jpg')
glDisable(GL_TEXTURE_2D)

-- i set up enable/disable to accept as many params as you can stuff in it
-- i know, i'm stretching my jurisdiction.
glEnable(GL_NORMALIZE, GL_CULL_FACE)

glLightModel{
	localViewer = true;
	ambient = {0,0,0,1};
	colorControl = GL_SEPARATE_SPECULAR_COLOR;
}

glLight{
	[0] = {		-- im debating better ways to reference which light # ... esp with lua 1-indexes
		ambient = {1,1,1,1};
		diffuse = {1,1,1,1};
		specular = {1,1,1,1};
		position = {0,0,1,0};
	}
}

glMaterial{
	front = {
		emission = {0,0,0,0};
		ambient = {0,0,0,1};
		diffuse = {1,1,1,1};
		specular = {1,1,1,1};
		shininess = 10;
	}
}

cowObj = loadOBJ('cow.obj')

-- make your fbo
fboID = glGenFramebuffer()
glBindFramebuffer(GL_FRAMEBUFFER, fboID)

-- make a depth buffer render target
-- do you need a depth target to do depth testing when rendering to a fbo?
fboDepthID = glGenRenderbuffer()
glBindRenderbuffer(GL_RENDERBUFFER, fboDepthID)
glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, fboSize, fboSize)
glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, fboDepthID)
glBindFramebuffer(GL_FRAMEBUFFER, 0)

-- make a tex 2D color buffer render target
-- make two of them.  two-state system.
fboTexs = {
	glGenTexture(),		-- two for advecting, cant rw to the same place 
	glGenTexture()		-- so we need two
}
for i,v in ipairs(fboTexs) do
	glBindTexture(GL_TEXTURE_2D, v)
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, fboSize, fboSize, 0, GL_RGBA, GL_UNSIGNED_BYTE)
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR) --GL_LINEAR_MIPMAP_LINEAR)
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
	-- autogen mipmapping isnt working for me up front
	--	glGenerateMipmap(GL_TEXTURE_2D)
	-- this is tho ... not really
	--	glTexParameter(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, true)
	--	glHint(GL_GENERATE_MIPMAP_HINT, GL_NICEST)
end

watercolorShader = createShader{
	vertexFile = 'watercolor.shader';
	vertexCode = defines('VERTEX_SHADER');
	fragmentFile = 'watercolor.shader';
	fragmentCode = defines(
		'FRAGMENT_SHADER',
		'TEX_SIZE '..fboSize..'.',
--		'TEX_SIZE 1000.',
		'ADVECT_STEP .0015'
	);
	uniforms={
		colorMap = 0;
		heightMap = 1;
	}
}
glUseProgramObject(0)
glActiveTexture(GL_TEXTURE0)
	
	
-- turn off clear on update ...
mainClearFlags(GL_DEPTH_BUFFER_BIT)

-- used in the loop
fboCycle = 1

first = true

-- callback from the main cheese
function display()
--[[
the watercolor gameplan
1. draw the cow
2. save it to a fbo
3. render the fbo to the whole screen
BUT WITH A SHADER!!

so how do we do accumulation?
--]]

	
	-- whats a good fractal pattern?
	-- render the thing as it is
	-- then texture it with an interpolation between itself and the fbo ...
	
	-- modulate back and forth
	local useFBO = true --math.fract(timer() * .25) < .5

	-- DO THIS EVERY TIME YOU WANT TO UPDATE THE TEXTURE:
	if useFBO then
	
		-- first render the ordinary scene to the third fbo tex
		-- hey do you need more than one fbo, or can you just keep rebinding it?
		-- so long as all the texs have the same size?
	
		glBindFramebuffer(GL_FRAMEBUFFER, fboID)
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, fboTexs[3], 0)
		local fboStatus = glCheckFramebufferStatus(GL_FRAMEBUFFER)
		if fboStatus ~= GL_FRAMEBUFFER_COMPLETE then
			print('got a bad fboStatus: '..fboStatus)
			return
		end

		glPushAttrib(GL_VIEWPORT_BIT)
		glViewport(0,0,fboSize,fboSize)
		--you dont need to reset the projection do you?
		
		-- render the static scene to the third fbo tex
		do
			--clear the buffer
			glClear(GL_COLOR_BUFFER_BIT)
-- [[
			-- draw our background
			glMatrixMode(GL_PROJECTION)
			glPushMatrix()
			glLoadIdentity()
			glOrtho(0,1,0,1,-1,1)
			glMatrixMode(GL_MODELVIEW)
			glPushMatrix()
			glLoadIdentity()
			
			glEnable(GL_TEXTURE_2D)
			glBindTexture(GL_TEXTURE_2D, backgroundTex)
			glColor(1,1,1)
			glBegin(GL_QUADS)
			glTexCoord(0,1)	glVertex(0,0)
			glTexCoord(1,1)	glVertex(1,0)
			glTexCoord(1,0)	glVertex(1,1)
			glTexCoord(0,0)	glVertex(0,1)
			glEnd()
			glDisable(GL_TEXTURE_2D)
			
			glMatrixMode(GL_PROJECTION)
			glPopMatrix()
			glMatrixMode(GL_MODELVIEW)
			glPopMatrix()
--]]
			-- clear the depth buffer
			glClear(GL_DEPTH_BUFFER_BIT)
			
			-- and draw the cow over it
-- [[
			glEnable(GL_LIGHTING, GL_LIGHT0, GL_DEPTH_TEST)
			drawOBJ(cowObj)
			glDisable(GL_LIGHTING, GL_LIGHT0, GL_DEPTH_TEST)
--]]
		end
		
		-- do i need to turn off the fbo if i'm gonna switch its texture?
		-- maybe i should  make a second one?
		glBindFramebuffer(GL_FRAMEBUFFER, 0)
		
		local fboNextCycle = 3 - fboCycle

		-- and bind back.  will it be done writing by now?
		glBindFramebuffer(GL_FRAMEBUFFER, fboID)
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, fboTexs[fboNextCycle], 0)
		local fboStatus = glCheckFramebufferStatus(GL_FRAMEBUFFER)
		if fboStatus ~= GL_FRAMEBUFFER_COMPLETE then
			print('got a bad fboStatus: '..fboStatus)
			return
		end

		-- draw scene here
		do			
			glMatrixMode(GL_PROJECTION)
			glLoadIdentity()
			glOrtho(0,1,0,1,-1,1)
			glMatrixMode(GL_MODELVIEW)
			glLoadIdentity()
			
			glClear(GL_COLOR_BUFFER_BIT)

			--draw our current gig
			glEnable(GL_BLEND)
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
			glColor(1,1,1, blendCoeff)
			glUseProgramObject(watercolorShader)

			glActiveTexture(GL_TEXTURE0)	--texture0 is the colormap
			glBindTexture(GL_TEXTURE_2D, fboTexs[fboNextCycle])
			glEnable(GL_TEXTURE_2D)
			glActiveTexture(GL_TEXTURE1)	--texture1 is the edgemap src
			glBindTexture(GL_TEXTURE_2D, backgroundTex)
			glEnable(GL_TEXTURE_2D)

			glBegin(GL_QUADS)
			glVertex(0,0)
			glVertex(1,0)
			glVertex(1,1)
			glVertex(0,1)
			glEnd()
			
			glActiveTexture(GL_TEXTURE1)
			glDisable(GL_TEXTURE_2D)
			glActiveTexture(GL_TEXTURE0)
			glDisable(GL_TEXTURE_2D)
			
			glUseProgramObject(0)
			glDisable(GL_BLEND)
			
			glColor(1,1,1,1)
		end

		glBindFramebuffer(GL_FRAMEBUFFER, 0)

		glPopAttrib()
		
		fboCycle = fboNextCycle
	end

	glClear(GL_DEPTH_BUFFER_BIT)
	glMatrixMode(GL_PROJECTION)
	glLoadIdentity()
	glOrtho(0,1,0,1,-1,1)
	
	glMatrixMode(GL_MODELVIEW)
	glLoadIdentity()

	glEnable(GL_TEXTURE_2D)
	glBindTexture(GL_TEXTURE_2D, fboTexs[fboCycle])

	glBegin(GL_QUADS)
	glTexCoord(0,0)	glVertex(0,0)
	glTexCoord(1,0)	glVertex(1,0)
	glTexCoord(1,1)	glVertex(1,1)
	glTexCoord(0,1)	glVertex(0,1)
	glEnd()

	glDisable(GL_TEXTURE_2D)
end
